@@ -2,7 +2,17 @@
 #   base.py                                                                   #
 #                                                                             #
 #   Copyright © 2015-2021, Rajiv Bakulesh Shah, original author.              #
-#   All rights reserved.                                                      #
+#                                                                             #
+#   Licensed under the Apache License, Version 2.0 (the "License");           #
+#   you may not use this file except in compliance with the License.          #
+#   You may obtain a copy of the License at:                                  #
+#       http://www.apache.org/licenses/LICENSE-2.0                            #
+#                                                                             #
+#   Unless required by applicable law or agreed to in writing, software       #
+#   distributed under the License is distributed on an "AS IS" BASIS,         #
+#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
+#   See the License for the specific language governing permissions and       #
+#   limitations under the License.                                            #
 # --------------------------------------------------------------------------- #
 
 
@@ -30,15 +40,18 @@ from typing import Tuple
 from typing import Type
 from typing import Union
 from typing import cast
+from typing import overload
 
 from redis import Redis
 from redis import RedisError
 from redis.client import Pipeline
 from typing_extensions import Final
+from typing_extensions import Literal
 
 from . import monkey
 from .annotations import JSONTypes
 from .annotations import RedisValues
+from .exceptions import QuorumIsImpossible
 from .exceptions import RandomKeyError
 
 
@@ -157,17 +170,16 @@ class _Comparable(metaclass=abc.ABCMeta):
 
     def __eq__(self, other: Any) -> bool:
         if self is other:
-            equals = True
-        elif (
+            return True
+        if (
             isinstance(other, _Comparable)
             and self.redis.connection_pool == other.redis.connection_pool  # NoQA: W503
             and self.key == other.key  # NoQA: W503
         ):
-            equals = True
-        else:
-            equals = super().__eq__(other)
-            if equals is NotImplemented:
-                equals = False
+            return True
+        equals = super().__eq__(other)
+        if equals is NotImplemented:
+            equals = False
         return equals
 
 
@@ -196,6 +208,8 @@ class _ContextPipeline:
     Redis transactions: https://redis.io/topics/transactions
     '''
 
+    __slots__ = ('redis', 'pipeline')
+
     def __init__(self, redis: Redis) -> None:
         self.redis = redis
 
@@ -203,16 +217,33 @@ class _ContextPipeline:
         self.pipeline = self.redis.pipeline()
         return self.pipeline
 
+    @overload
+    def __exit__(self,
+                 exc_type: None,
+                 exc_value: None,
+                 exc_traceback: None,
+                 ) -> Literal[False]:
+        raise NotImplementedError
+
+    @overload
+    def __exit__(self,
+                 exc_type: Type[BaseException],
+                 exc_value: BaseException,
+                 exc_traceback: TracebackType,
+                 ) -> Literal[False]:
+        raise NotImplementedError
+
     def __exit__(self,
                  exc_type: Optional[Type[BaseException]],
                  exc_value: Optional[BaseException],
-                 traceback: Optional[TracebackType],
-                 ) -> None:
+                 exc_traceback: Optional[TracebackType],
+                 ) -> Literal[False]:
         if exc_type is None:
             with contextlib.suppress(RedisError):
                 self.pipeline.multi()
                 self.pipeline.ping()
             self.pipeline.execute()
+        return False
 
 
 class _Pipelined(metaclass=abc.ABCMeta):
@@ -279,7 +310,8 @@ class Iterable_(metaclass=abc.ABCMeta):
 
     @abc.abstractmethod
     def _scan(self,
-              *, cursor: int = 0,
+              *,
+              cursor: int = 0,
               ) -> Tuple[int, Union[List[bytes], Mapping[bytes, bytes]]]:
         raise NotImplementedError
 
@@ -296,15 +328,19 @@ class Iterable_(metaclass=abc.ABCMeta):
 class Primitive(metaclass=abc.ABCMeta):
     'Base class for Redis-backed distributed primitives.'
 
+    __slots__ = ('_key', 'masters', 'raise_on_redis_errors')
+
     _DEFAULT_MASTERS: ClassVar[FrozenSet[Redis]] = frozenset({_default_redis})
 
     def __init__(self,
                  *,
                  key: str,
                  masters: Iterable[Redis] = frozenset(),
+                 raise_on_redis_errors: bool = False,
                  ) -> None:
         self.key = key
         self.masters = frozenset(masters) or self._DEFAULT_MASTERS
+        self.raise_on_redis_errors = raise_on_redis_errors
 
     @property
     @abc.abstractmethod
@@ -318,3 +354,16 @@ class Primitive(metaclass=abc.ABCMeta):
     @key.setter
     def key(self, value: str) -> None:
         self._key = f'{self.KEY_PREFIX}:{value}'
+
+    def _check_enough_masters_up(self,
+                                 raise_on_redis_errors: Optional[bool],
+                                 redis_errors: List[RedisError],
+                                 ) -> None:
+        if raise_on_redis_errors is None:
+            raise_on_redis_errors = self.raise_on_redis_errors
+        if raise_on_redis_errors and len(redis_errors) > len(self.masters) // 2:
+            raise QuorumIsImpossible(
+                self.key,
+                self.masters,
+                redis_errors=redis_errors,
+            )
