@@ -2,7 +2,17 @@
 #   redlock.py                                                                #
 #                                                                             #
 #   Copyright © 2015-2021, Rajiv Bakulesh Shah, original author.              #
-#   All rights reserved.                                                      #
+#                                                                             #
+#   Licensed under the Apache License, Version 2.0 (the "License");           #
+#   you may not use this file except in compliance with the License.          #
+#   You may obtain a copy of the License at:                                  #
+#       http://www.apache.org/licenses/LICENSE-2.0                            #
+#                                                                             #
+#   Unless required by applicable law or agreed to in writing, software       #
+#   distributed under the License is distributed on an "AS IS" BASIS,         #
+#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
+#   See the License for the specific language governing permissions and       #
+#   limitations under the License.                                            #
 # --------------------------------------------------------------------------- #
 '''Distributed Redis-powered lock.
 
@@ -26,26 +36,31 @@ import concurrent.futures
 import contextlib
 import functools
 import logging
-import os
+import math
 import random
 import time
+import uuid
 from types import TracebackType
 from typing import Any
 from typing import Callable
 from typing import ClassVar
 from typing import Iterable
 from typing import Optional
+from typing import Tuple
 from typing import Type
 from typing import cast
+from typing import overload
 
 from redis import Redis
 from redis import RedisError
 from redis.client import Script
 from typing_extensions import Final
+from typing_extensions import Literal
 
 from .annotations import F
 from .base import Primitive
 from .exceptions import ExtendUnlockedLock
+from .exceptions import QuorumNotAchieved
 from .exceptions import ReleaseUnlockedLock
 from .exceptions import TooManyExtensions
 from .executor import BailOutExecutor
@@ -58,7 +73,79 @@ AUTO_RELEASE_TIME: Final[int] = 10 * 1000
 _logger: Final[logging.Logger] = logging.getLogger('pottery')
 
 
-class Redlock(Primitive):
+class _Scripts:
+    '''Parent class to define/register Lua scripts for Redis.
+
+    Note that we only have to register these Lua scripts once -- so we do it on
+    the first instantiation of Redlock.
+    '''
+
+    __slots__: Tuple[str, ...] = tuple()
+
+    _acquired_script: ClassVar[Optional[Script]] = None
+    _extend_script: ClassVar[Optional[Script]] = None
+    _release_script: ClassVar[Optional[Script]] = None
+
+    def __init__(self,
+                 *,
+                 key: str,
+                 masters: Iterable[Redis] = frozenset(),
+                 raise_on_redis_errors: bool = False,
+                 ) -> None:
+        super().__init__(  # type: ignore
+            key=key,
+            masters=masters,
+            raise_on_redis_errors=raise_on_redis_errors,
+        )
+        self.__register_acquired_script()
+        self.__register_extend_script()
+        self.__register_release_script()
+
+    # Preserve the Open-Closed Principle with name mangling.
+    #   https://youtu.be/miGolgp9xq8?t=2086
+    #   https://stackoverflow.com/a/38534939
+    def __register_acquired_script(self) -> None:
+        if self._acquired_script is None:
+            class_name = self.__class__.__qualname__
+            _logger.info('Registering %s._acquired_script', class_name)
+            master = next(iter(self.masters))  # type: ignore
+            self.__class__._acquired_script = master.register_script('''
+                if redis.call('get', KEYS[1]) == ARGV[1] then
+                    local pttl = redis.call('pttl', KEYS[1])
+                    return (pttl > 0) and pttl or 0
+                else
+                    return 0
+                end
+            ''')
+
+    def __register_extend_script(self) -> None:
+        if self._extend_script is None:
+            class_name = self.__class__.__qualname__
+            _logger.info('Registering %s._extend_script', class_name)
+            master = next(iter(self.masters))  # type: ignore
+            self.__class__._extend_script = master.register_script('''
+                if redis.call('get', KEYS[1]) == ARGV[1] then
+                    return redis.call('pexpire', KEYS[1], ARGV[2])
+                else
+                    return 0
+                end
+            ''')
+
+    def __register_release_script(self) -> None:
+        if self._release_script is None:
+            class_name = self.__class__.__qualname__
+            _logger.info('Registering %s._release_script', class_name)
+            master = next(iter(self.masters))  # type: ignore
+            self.__class__._release_script = master.register_script('''
+                if redis.call('get', KEYS[1]) == ARGV[1] then
+                    return redis.call('del', KEYS[1])
+                else
+                    return 0
+                end
+            ''')
+
+
+class Redlock(_Scripts, Primitive):
     '''Distributed Redis-powered lock.
 
     This algorithm safely and reliably provides a mutually-exclusive locking
@@ -133,99 +220,59 @@ class Redlock(Primitive):
         [True, False]
     '''
 
+    __slots__ = (
+        'auto_release_time',
+        'num_extensions',
+        'context_manager_blocking',
+        'context_manager_timeout',
+        '_uuid',
+        '_extension_num',
+    )
+
     KEY_PREFIX: ClassVar[str] = 'redlock'
     CLOCK_DRIFT_FACTOR: ClassVar[float] = 0.01
     RETRY_DELAY: ClassVar[int] = 200
     NUM_EXTENSIONS: ClassVar[int] = 3
-    NUM_RANDOM_BYTES: ClassVar[int] = 20
-
-    _acquired_script: ClassVar[Optional[Script]] = None
-    _extend_script: ClassVar[Optional[Script]] = None
-    _release_script: ClassVar[Optional[Script]] = None
 
     def __init__(self,
                  *,
                  key: str,
                  masters: Iterable[Redis] = frozenset(),
+                 raise_on_redis_errors: bool = False,
                  auto_release_time: int = AUTO_RELEASE_TIME,
                  num_extensions: int = NUM_EXTENSIONS,
-                 num_random_bytes: int = NUM_RANDOM_BYTES,
+                 context_manager_blocking: bool = True,
+                 context_manager_timeout: float = -1,
                  ) -> None:
-        super().__init__(key=key, masters=masters)
-        self.__register_acquired_script()
-        self.__register_extend_script()
-        self.__register_release_script()
+        if not context_manager_blocking and context_manager_timeout != -1:
+            raise ValueError("can't specify a timeout for a non-blocking call")
 
+        super().__init__(
+            key=key,
+            masters=masters,
+            raise_on_redis_errors=raise_on_redis_errors,
+        )
         self.auto_release_time = auto_release_time
         self.num_extensions = num_extensions
-        self.num_random_bytes = num_random_bytes
-
-        self._value = b''
+        self.context_manager_blocking = context_manager_blocking
+        self.context_manager_timeout = context_manager_timeout
+        self._uuid = ''
         self._extension_num = 0
 
-    # Preserve the Open-Closed Principle with name mangling.
-    #   https://youtu.be/miGolgp9xq8?t=2086
-    #   https://stackoverflow.com/a/38534939
-    def __register_acquired_script(self) -> None:
-        if self._acquired_script is None:
-            _logger.info(
-                'Registering %s._acquired_script',
-                self.__class__.__name__,
-            )
-            master = next(iter(self.masters))
-            self.__class__._acquired_script = master.register_script('''
-                if redis.call('get', KEYS[1]) == ARGV[1] then
-                    local pttl = redis.call('pttl', KEYS[1])
-                    return (pttl > 0) and pttl or 0
-                else
-                    return 0
-                end
-            ''')
-
-    def __register_extend_script(self) -> None:
-        if self._extend_script is None:
-            _logger.info(
-                'Registering %s._extend_script',
-                self.__class__.__name__,
-            )
-            master = next(iter(self.masters))
-            self.__class__._extend_script = master.register_script('''
-                if redis.call('get', KEYS[1]) == ARGV[1] then
-                    return redis.call('pexpire', KEYS[1], ARGV[2])
-                else
-                    return 0
-                end
-            ''')
-
-    def __register_release_script(self) -> None:
-        if self._release_script is None:
-            _logger.info(
-                'Registering %s._release_script',
-                self.__class__.__name__,
-            )
-            master = next(iter(self.masters))
-            self.__class__._release_script = master.register_script('''
-                if redis.call('get', KEYS[1]) == ARGV[1] then
-                    return redis.call('del', KEYS[1])
-                else
-                    return 0
-                end
-            ''')
-
     def __acquire_master(self, master: Redis) -> bool:
         acquired = master.set(
             self.key,
-            self._value,
+            self._uuid,
             px=self.auto_release_time,
             nx=True,
         )
         return bool(acquired)
 
     def __acquired_master(self, master: Redis) -> int:
-        if self._value:
+        if self._uuid:
             ttl: int = cast(Script, self._acquired_script)(
                 keys=(self.key,),
-                args=(self._value,),
+                args=(self._uuid,),
                 client=master,
             )
         else:
@@ -235,7 +282,7 @@ class Redlock(Primitive):
     def __extend_master(self, master: Redis) -> bool:
         extended = cast(Script, self._extend_script)(
             keys=(self.key,),
-            args=(self._value, self.auto_release_time),
+            args=(self._uuid, self.auto_release_time),
             client=master,
         )
         return bool(extended)
@@ -243,7 +290,7 @@ class Redlock(Primitive):
     def __release_master(self, master: Redis) -> bool:
         released = cast(Script, self._release_script)(
             keys=(self.key,),
-            args=(self._value,),
+            args=(self._uuid,),
             client=master,
         )
         return bool(released)
@@ -251,10 +298,12 @@ class Redlock(Primitive):
     def __drift(self) -> float:
         return self.auto_release_time * self.CLOCK_DRIFT_FACTOR + 2
 
-    def __acquire_masters(self) -> bool:
-        self._value = os.urandom(self.num_random_bytes)
+    def __acquire_masters(self,
+                          *,
+                          raise_on_redis_errors: Optional[bool] = None,
+                          ) -> bool:
+        self._uuid = str(uuid.uuid4())
         self._extension_num = 0
-        quorum, validity_time = False, 0.0
 
         with ContextTimer() as timer, BailOutExecutor() as executor:
             futures = set()
@@ -262,31 +311,36 @@ class Redlock(Primitive):
                 future = executor.submit(self.__acquire_master, master)
                 futures.add(future)
 
-            num_masters_acquired = 0
+            num_masters_acquired, redis_errors = 0, []
             for future in concurrent.futures.as_completed(futures):
                 try:
                     num_masters_acquired += future.result()
-                except RedisError as error:  # pragma: no cover
+                except RedisError as error:
+                    redis_errors.append(error)
                     _logger.exception(
-                        '%s.__acquire_masters() caught an %s',
+                        '%s.__acquire_masters() caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
                     )
                 else:
-                    quorum = num_masters_acquired >= len(self.masters) // 2 + 1
-                    if quorum:
-                        elapsed = timer.elapsed() - self.__drift()
-                        validity_time = self.auto_release_time - elapsed
-                        break
-
-        if quorum and max(validity_time, 0):
-            return True
+                    if num_masters_acquired > len(self.masters) // 2:
+                        validity_time = self.auto_release_time
+                        validity_time -= round(self.__drift())
+                        validity_time -= timer.elapsed()
+                        if validity_time > 0:  # pragma: no cover
+                            return True
 
         with contextlib.suppress(ReleaseUnlockedLock):
-            self.__release()
+            self.__release(raise_on_redis_errors=False)
+        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
         return False
 
-    def acquire(self, *, blocking: bool = True, timeout: int = -1) -> bool:
+    def acquire(self,
+                *,
+                blocking: bool = True,
+                timeout: float = -1,
+                raise_on_redis_errors: Optional[bool] = None,
+                ) -> bool:
         '''Lock the lock.
 
         If blocking is True and timeout is -1, then wait for as long as
@@ -330,22 +384,44 @@ class Redlock(Primitive):
             False
             >>> printer_lock_1.release()
         '''
+        acquire_masters = functools.partial(
+            self.__acquire_masters,
+            raise_on_redis_errors=raise_on_redis_errors,
+        )
+
+        def log_time_enqueued(timer: ContextTimer, acquired: bool) -> None:
+            key_suffix = self.key.split(':', maxsplit=1)[1]
+            time_enqueued = math.ceil(timer.elapsed())
+            _logger.info(
+                'source=pottery sample#redlock.enqueued.%s=%dms sample#redlock.acquired.%s=%d',
+                key_suffix,
+                time_enqueued,
+                key_suffix,
+                acquired,
+            )
+
         if blocking:
+            enqueued = False
             with ContextTimer() as timer:
                 while timeout == -1 or timer.elapsed() / 1000 < timeout:
-                    if self.__acquire_masters():
+                    if acquire_masters():
+                        if enqueued:
+                            log_time_enqueued(timer, True)
                         return True
+                    enqueued = True
                     time.sleep(random.uniform(0, self.RETRY_DELAY/1000))
+            if enqueued:
+                log_time_enqueued(timer, False)
             return False
 
         if timeout == -1:
-            return self.__acquire_masters()
+            return acquire_masters()
 
         raise ValueError("can't specify a timeout for a non-blocking call")
 
     __acquire = acquire
 
-    def locked(self) -> int:
+    def locked(self, *, raise_on_redis_errors: Optional[bool] = None) -> int:
         '''How much longer we'll hold the lock (unless we extend or release it).
 
         If we don't currently hold the lock, then this method returns 0.
@@ -371,37 +447,38 @@ class Redlock(Primitive):
             >>> printer_lock_1.release()
 
         '''
-        with ContextTimer() as timer, \
-             concurrent.futures.ThreadPoolExecutor() as executor:
+        with ContextTimer() as timer, BailOutExecutor() as executor:
             futures = set()
             for master in self.masters:
                 future = executor.submit(self.__acquired_master, master)
                 futures.add(future)
 
-            ttls = []
+            ttls, redis_errors = [], []
             for future in concurrent.futures.as_completed(futures):
                 try:
-                    ttls.append(future.result())
-                except RedisError as error:  # pragma: no cover
+                    ttl = future.result()
+                except RedisError as error:
+                    redis_errors.append(error)
                     _logger.exception(
-                        '%s.locked() caught an %s',
+                        '%s.locked() caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
                     )
+                else:
+                    if ttl:
+                        ttls.append(ttl)
+                        if len(ttls) > len(self.masters) // 2:  # pragma: no cover
+                            validity_time = min(ttls)
+                            validity_time -= round(self.__drift())
+                            validity_time -= timer.elapsed()
+                            return max(validity_time, 0)
 
-            num_masters_acquired = sum(1 for ttl in ttls if ttl > 0)
-            quorum = num_masters_acquired >= len(self.masters) // 2 + 1
-            if quorum:
-                ttls = sorted(ttls, reverse=True)
-                validity_time = ttls[len(self.masters) // 2]
-                validity_time -= round(timer.elapsed() + self.__drift())
-                return max(validity_time, 0)
-
-            return 0
+        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
+        return 0
 
     __locked = locked
 
-    def extend(self) -> None:
+    def extend(self, *, raise_on_redis_errors: Optional[bool] = None) -> None:
         '''Extend our hold on the lock (if we currently hold it).
 
         Usage:
@@ -421,7 +498,6 @@ class Redlock(Primitive):
         '''
         if self._extension_num >= self.num_extensions:
             raise TooManyExtensions(self.key, self.masters)
-        quorum = False
 
         with BailOutExecutor() as executor:
             futures = set()
@@ -429,26 +505,30 @@ class Redlock(Primitive):
                 future = executor.submit(self.__extend_master, master)
                 futures.add(future)
 
-            num_masters_extended = 0
+            num_masters_extended, redis_errors = 0, []
             for future in concurrent.futures.as_completed(futures):
                 try:
                     num_masters_extended += future.result()
-                except RedisError as error:  # pragma: no cover
+                except RedisError as error:
+                    redis_errors.append(error)
                     _logger.exception(
-                        '%s.extend() caught an %s',
+                        '%s.extend() caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
                     )
                 else:
-                    quorum = num_masters_extended >= len(self.masters) // 2 + 1
-                    if quorum:
-                        break
+                    if num_masters_extended > len(self.masters) // 2:
+                        self._extension_num += 1
+                        return
 
-        self._extension_num += quorum
-        if not quorum:
-            raise ExtendUnlockedLock(self.key, self.masters)
+        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
+        raise ExtendUnlockedLock(
+            self.key,
+            self.masters,
+            redis_errors=redis_errors,
+        )
 
-    def release(self) -> None:
+    def release(self, *, raise_on_redis_errors: Optional[bool] = None) -> None:
         '''Unlock the lock.
 
         Usage:
@@ -464,31 +544,33 @@ class Redlock(Primitive):
             >>> bool(printer_lock.locked())
             False
         '''
-        quorum = False
-
         with BailOutExecutor() as executor:
             futures = set()
             for master in self.masters:
                 future = executor.submit(self.__release_master, master)
                 futures.add(future)
 
-            num_masters_released = 0
+            num_masters_released, redis_errors = 0, []
             for future in concurrent.futures.as_completed(futures):
                 try:
                     num_masters_released += future.result()
-                except RedisError as error:  # pragma: no cover
+                except RedisError as error:
+                    redis_errors.append(error)
                     _logger.exception(
-                        '%s.release() caught an %s',
+                        '%s.release() caught %s',
                         self.__class__.__name__,
                         error.__class__.__name__,
                     )
                 else:
-                    quorum = num_masters_released >= len(self.masters) // 2 + 1
-                    if quorum:
-                        break
+                    if num_masters_released > len(self.masters) // 2:
+                        return
 
-        if not quorum:
-            raise ReleaseUnlockedLock(self.key, self.masters)
+        self._check_enough_masters_up(raise_on_redis_errors, redis_errors)
+        raise ReleaseUnlockedLock(
+            self.key,
+            self.masters,
+            redis_errors=redis_errors,
+        )
 
     __release = release
 
@@ -513,14 +595,35 @@ class Redlock(Primitive):
             >>> states
             [True, False]
         '''
-        self.__acquire()
-        return self
+        acquired = self.__acquire(
+            blocking=self.context_manager_blocking,
+            timeout=self.context_manager_timeout,
+        )
+        if acquired:
+            return self
+        raise QuorumNotAchieved(self.key, self.masters)
+
+    @overload
+    def __exit__(self,
+                 exc_type: None,
+                 exc_value: None,
+                 exc_traceback: None,
+                 ) -> Literal[False]:
+        raise NotImplementedError
+
+    @overload
+    def __exit__(self,
+                 exc_type: Type[BaseException],
+                 exc_value: BaseException,
+                 exc_traceback: TracebackType,
+                 ) -> Literal[False]:
+        raise NotImplementedError
 
     def __exit__(self,
                  exc_type: Optional[Type[BaseException]],
                  exc_value: Optional[BaseException],
                  traceback: Optional[TracebackType],
-                 ) -> None:
+                 ) -> Literal[False]:
         '''You can use a Redlock as a context manager.
 
         Usage:
@@ -542,18 +645,19 @@ class Redlock(Primitive):
             [True, False]
         '''
         self.__release()
+        return False
 
     def __repr__(self) -> str:
-        return (
-            f'<{self.__class__.__name__} key={self.key} '
-            f'value={str(self._value)} timeout={self.__locked()}>'
-        )
+        return f'<{self.__class__.__name__} key={self.key}>'
 
 
 def synchronize(*,
                 key: str,
                 masters: Iterable[Redis] = frozenset(),
+                raise_on_redis_errors: bool = False,
                 auto_release_time: int = AUTO_RELEASE_TIME,
+                blocking: bool = True,
+                timeout: float = -1,
                 ) -> Callable[[F], F]:
     '''Decorator to synchronize a function's execution across threads.
 
@@ -571,28 +675,62 @@ def synchronize(*,
         >>> func()
         True
     '''
+    RedlockFactory = functools.partial(
+        Redlock,
+        key=key,
+        masters=masters,
+        raise_on_redis_errors=raise_on_redis_errors,
+        auto_release_time=auto_release_time,
+        context_manager_blocking=blocking,
+        context_manager_timeout=timeout,
+    )
+
     def decorator(func: F) -> F:
         @functools.wraps(func)
         def wrapper(*args: Any, **kwargs: Any) -> Any:
-            redlock = Redlock(
-                key=key,
-                masters=masters,
-                auto_release_time=auto_release_time,
-            )
-            with ContextTimer() as timer, redlock:
-                return_value = func(*args, **kwargs)
+            redlock = RedlockFactory()
+            waiting_timer, holding_timer = ContextTimer(), ContextTimer()
+
+            try:
+                waiting_timer.start()
+                with redlock:
+                    waiting_timer.stop()
+                    holding_timer.start()
+                    return_value = func(*args, **kwargs)
+                holding_timer.stop()
+            finally:
+                _log_synchronize(func, redlock, waiting_timer, holding_timer)
 
-            _logger.info(
-                '%s() held %s for %d ms',
-                func.__qualname__,
-                redlock.key,
-                timer.elapsed(),
-            )
             return return_value
         return cast(F, wrapper)
     return decorator
 
 
+def _log_synchronize(func: F,
+                     redlock: Redlock,
+                     waiting_timer: ContextTimer,
+                     holding_timer: ContextTimer,
+                     ) -> None:
+    try:
+        _logger.info(
+            '%s() waited for %s for %d ms; held for %d ms',
+            func.__qualname__,
+            redlock.key,
+            waiting_timer.elapsed(),
+            holding_timer.elapsed(),
+        )
+    except RuntimeError:  # pragma: no cover
+        # holding_timer.elapsed() threw a RuntimeError, which means that
+        # holding_timer never started, which means that we never acquired the
+        # lock / entered the critical section.
+        _logger.info(
+            '%s() waited for %s for %d ms; never acquired lock',
+            func.__qualname__,
+            redlock.key,
+            waiting_timer.elapsed(),
+        )
+
+
 if __name__ == '__main__':
     # Run the doctests in this module with:
     #   $ source venv/bin/activate
