@@ -2,7 +2,17 @@
 #   set.py                                                                    #
 #                                                                             #
 #   Copyright © 2015-2021, Rajiv Bakulesh Shah, original author.              #
-#   All rights reserved.                                                      #
+#                                                                             #
+#   Licensed under the Apache License, Version 2.0 (the "License");           #
+#   you may not use this file except in compliance with the License.          #
+#   You may obtain a copy of the License at:                                  #
+#       http://www.apache.org/licenses/LICENSE-2.0                            #
+#                                                                             #
+#   Unless required by applicable law or agreed to in writing, software       #
+#   distributed under the License is distributed on an "AS IS" BASIS,         #
+#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
+#   See the License for the specific language governing permissions and       #
+#   limitations under the License.                                            #
 # --------------------------------------------------------------------------- #
 
 
@@ -14,6 +24,7 @@ from typing import List
 from typing import NoReturn
 from typing import Optional
 from typing import Tuple
+from typing import cast
 
 from redis import Redis
 from redis.client import Pipeline
@@ -89,7 +100,7 @@ class RedisSet(Base, Iterable_, collections.abc.MutableSet):
         encoded_value = self.redis.spop(self.key)
         if encoded_value is None:
             raise KeyError('pop from an empty set')
-        return self._decode(encoded_value)
+        return self._decode(cast(bytes, encoded_value))
 
     # From collections.abc.MutableSet:
     def remove(self, value: JSONTypes) -> None:
