@@ -2,7 +2,17 @@
 #   timer.py                                                                  #
 #                                                                             #
 #   Copyright © 2015-2021, Rajiv Bakulesh Shah, original author.              #
-#   All rights reserved.                                                      #
+#                                                                             #
+#   Licensed under the Apache License, Version 2.0 (the "License");           #
+#   you may not use this file except in compliance with the License.          #
+#   You may obtain a copy of the License at:                                  #
+#       http://www.apache.org/licenses/LICENSE-2.0                            #
+#                                                                             #
+#   Unless required by applicable law or agreed to in writing, software       #
+#   distributed under the License is distributed on an "AS IS" BASIS,         #
+#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  #
+#   See the License for the specific language governing permissions and       #
+#   limitations under the License.                                            #
 # --------------------------------------------------------------------------- #
 'Measure the execution time of small code snippets.'
 
@@ -11,6 +21,9 @@ import timeit
 from types import TracebackType
 from typing import Optional
 from typing import Type
+from typing import overload
+
+from typing_extensions import Literal
 
 
 class ContextTimer:
@@ -54,12 +67,29 @@ class ContextTimer:
         self.__start()
         return self
 
+    @overload
+    def __exit__(self,
+                 exc_type: None,
+                 exc_value: None,
+                 exc_traceback: None,
+                 ) -> Literal[False]:
+        raise NotImplementedError
+
+    @overload
+    def __exit__(self,
+                 exc_type: Type[BaseException],
+                 exc_value: BaseException,
+                 exc_traceback: TracebackType,
+                 ) -> Literal[False]:
+        raise NotImplementedError
+
     def __exit__(self,
                  exc_type: Optional[Type[BaseException]],
                  exc_value: Optional[BaseException],
                  exc_traceback: Optional[TracebackType],
-                 ) -> None:
+                 ) -> Literal[False]:
         self.__stop()
+        return False
 
     def start(self) -> None:
         if self._stopped:
